/* -*- mode: C; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 * Copyright (c) 2014, Tbricks AB
 * Copyright (c) 2018 Samsung Electronics Co., Ltd.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL TBRICKS
 * AB BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THISS OFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include "mem_map.h"
#include "proc.h"

#include <ctype.h>
#include <dirent.h>
#include <elf.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <linux/limits.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/syscall.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/wait.h>
#include <unistd.h>

#define LOG_TAG "CRASH_STACK"
#include "shared/log.h"

#ifndef SYS_process_vm_readv
#if defined(__i386)
#define SYS_process_vm_readv 365
#elif defined(__x86_64)
#define SYS_process_vm_readv 310
#elif defined(__arm__) || defined(__aarch64__)
#define SYS_process_vm_readv 376
#else
#error SYS_process_vm_readv is undefined
#endif
#endif

#define SLEEP_WAIT 500

int attached_pid = 0;

/* timeout on waiting for process to stop (us) */
extern int stop_timeout;
/* static int sleep_time = 0; */

/* for summary */
int sleep_count = 0;
size_t total_length = 0;

extern struct timeval freeze_time;
extern struct timeval unfreeze_time;

/* extern int opt_proc_mem; */
/* extern int opt_use_waitpid_timeout; */
/* extern int opt_verbose; */

static int opt_proc_mem = 0;
static int opt_verbose = 0;

int proc_state(int pid)
{
    FILE *f;
    char buf[128];
    char c;
    int res = -1;

    snprintf(buf, sizeof(buf), "/proc/%d/status", pid);
    if ((f = fopen(buf, "r")) == NULL) {
        _E("Cannot open %s: %s", buf, strerror(errno));
        return -1;
    }

    while (fgets(buf, sizeof(buf), f)) {
        if (sscanf(buf, "State:\t%c", &c) == 1) {
            res = c;
            break;
        }
    }

    fclose(f);
    return res;
}

struct mem_map *create_maps(int pid)
{
    FILE *f = NULL;
    char *buf = NULL, *str = NULL;
    size_t total_read, capacity, count;

    size_t addr_start, addr_end, offset, len;
    char r, w, x, p;
    int dev_major, dev_minor, inode;
    char path[PATH_MAX];

    struct mem_map *map = NULL;
    struct mem_region *region;

    capacity = 0x100000;
    buf = calloc(1, capacity);
    if (buf == NULL) {
        _E("Unable to allocate memory: %s", strerror(errno));
        return NULL;
    }
    snprintf(buf, capacity, "/proc/%d/maps", pid);
    if ((f = fopen(buf, "r")) == NULL) {
        _E("Cannot open %s: %s", buf, strerror(errno));
        goto create_maps_end;
    }

    map = malloc(sizeof(struct mem_map));
    if (map == NULL){
        _E("Unable to allocate memory: %s",  strerror(errno));
        goto create_maps_end;
    }
    mem_map_init(map);

    memset(buf, 0, capacity);
    total_read = 0;
    while (!feof(f)) {
        count = fread(&buf[total_read], capacity - total_read - 1, 1, f);
        if (count <1 && ferror(f)) {
            clearerr(f);
            perror("maps");
            mem_map_destroy(map);
            map = NULL;
            goto create_maps_end;
        }

        total_read = strlen(buf);
        if ((total_read + 1) == capacity) {
            capacity *= 2;
            buf = realloc(buf, capacity);
            if (buf == NULL) {
                _E("Unable to reallocate memory: %s", strerror(errno));
                mem_map_destroy(map);
                map = NULL;
                goto create_maps_end;
            }
            memset(&buf[total_read], 0, capacity - total_read);
        } else {
            buf[total_read] = '\0';
        }
    }

    str = &buf[0];
    while (*str) {
        int scan;
        char *next;

        next = strchr(str, '\n');
        if (next != NULL)
            *next = '\0';

        scan = sscanf(str, "%zx-%zx %c%c%c%c %zx %x:%x %d %[^\t\n]",
                &addr_start, &addr_end,
                &r, &w, &x, &p,
                &offset,
                &dev_major, &dev_minor,
                &inode,
                path);

        if (scan < 10) {
            _W("Unable to parse maps "
                    "entry '%s' (read %d)", str, scan);
            break;
        }

        region = malloc(sizeof(struct mem_region));
        if (region == NULL) {
            _E("Unable to allocate memory: %s", strerror(errno));
            mem_map_destroy(map);
            map = NULL;
            break;
        }
        mem_region_init(region);

        region->start = (void *)addr_start;
        region->length = addr_end - addr_start;
        region->offset = offset;
        if (scan > 10 && path[0] != '\0') {
            if (!strcmp(path, "[vdso]")) {
                region->type = MEM_REGION_TYPE_VDSO;
            } else if (!strcmp(path, "[vsyscall]")) {
                region->type = MEM_REGION_TYPE_VSYSCALL;
            } else if ((len = strlen(path)) > 10 &&
                    !strcmp(path + len - 10, " (deleted)")) {
                *(path + len - 10) = '\0';
                region->path = strdup(path);
                region->type = MEM_REGION_TYPE_DELETED;
            } else {
                region->path = strdup(path);
                region->type = MEM_REGION_TYPE_MMAP;
            }
        }

        if (mem_map_add_region(map, region) != 0) {
            mem_region_destroy(region);
            mem_map_destroy(map);
            map = NULL;
            break;
        }

        if (next != NULL)
            str = next + 1;
    }

    if (map != NULL)
        mem_map_create_region_index(map);

create_maps_end:
    if (f != NULL)
        fclose(f);
    if (buf != NULL)
        free(buf);
    return map;
}

int print_proc_maps(int pid)
{
    char cmd[32];
    snprintf(cmd, sizeof(cmd), "cat /proc/%d/maps 1>&2", pid);
    return system(cmd);
}

/*
 * filter for scandir(). choose only thread identifiers
 */
static int dir_select(const struct dirent *entry)
{
    const char *c = entry->d_name;
    while (*c)
        if (!isdigit(*c++))
            return 0;
    return 1;
}

int get_threads(int pid, int **tids)
{
    char buf[32];
    struct dirent **namelist;
    int cur, i, n;

    snprintf(buf, sizeof(buf), "/proc/%d/task", pid);

    n = scandir(buf, &namelist, dir_select, NULL);
    if (n < 0) {
        perror(buf);
        return -1;
    } else {
        *tids = malloc(sizeof(int)*n);
        i = 0;
        while (i < n) {
            cur = atoi(namelist[i]->d_name);
            (*tids)[i] = cur;
            free(namelist[i++]);
        }
        free(namelist);
    }

    return n;
}

char *get_thread_states(const int *tids, int n)
{
    int i;
    char *res = calloc(1, n);

    for (i = 0; i < n; ++i) {
        int state = proc_state(tids[i]);
        if (state < 0) {
            _W("Could not get state of thread %d",
                    tids[i]);
            res[i] = '?';
            continue;
        }
        res[i] = state;
    }

    return res;
}

int adjust_threads(int *tids, int nr_tids, int *user_tids,
        int *index, int nr_user)
{
    int i, j, n = 0;
    for (i = 0; i < nr_user; ++i) {
        int found = 0;
        for (j = 0; j < nr_tids; ++j) {
            if (tids[j] == user_tids[i]) {
                found = 1;
                break;
            }
        }
        if (!found) {
            if (n || (user_tids[i] > nr_tids) || (user_tids[i] <= 0)) {
                _E("Unexpected thread %d", user_tids[i]);
                return -1;
            }
        } else {
            ++n;
            index[i] = j + 1;
        }
    }
    if (!n) {
        for (i = 0; i < nr_user; ++i) {
            index[i] = user_tids[i];
            user_tids[i] = tids[user_tids[i]-1];
        }
    }
    return 0;
}

int filter_threads(int tids[], int index[], char states[], int nr_tids,
        const char *user_states)
{
    int i, j = 0;
    int nr_prev = nr_tids;
    for (i = 0; i < nr_prev; ++i) {
        if (strchr(user_states, states[i]) != NULL) {
            if (i != j) {
                tids[j] = tids[i];
                states[j] = states[i];
                if (index != NULL)
                    index[j] = index[i];
            }
            ++j;
        } else {
            --nr_tids;
        }
    }
    return nr_tids;
}

/*
 * copy memory contents using process_vm_readv(). reduces number
 * of system calls comparing to /proc/pid/mem
 *
 * return values:
 *      0  success
 *     -1  fail
 * ENOSYS  process_vm_readv() is not supported
 */
static int copy_memory_process_vm_readv(int pid,
        struct mem_data_chunk **frames, int n_frames)
{
    struct iovec *local_iov, *remote_iov;
    ssize_t *frame_bytes;
    int i, rc = -1;
    ssize_t bytes_total = 0;
    int seg_count = 0;

    local_iov = malloc(sizeof(struct iovec)*n_frames);
    remote_iov = malloc(sizeof(struct iovec)*n_frames);
    frame_bytes = malloc(sizeof(ssize_t)*n_frames);

    for (i = 0; i < n_frames; ++i) {
        local_iov[i].iov_base = frames[i]->data;
        local_iov[i].iov_len = frames[i]->length;
        remote_iov[i].iov_base = frames[i]->start;
        remote_iov[i].iov_len = frames[i]->length;

        bytes_total += frames[i]->length;
        frame_bytes[i] = bytes_total;
    }

    bytes_total = 0;
    while (1) {
        ssize_t bytes_read;
        int frames_to_read = n_frames - seg_count;
        if (frames_to_read > IOV_MAX)
            frames_to_read = IOV_MAX;

        bytes_read = syscall(SYS_process_vm_readv,
                pid,
                local_iov + seg_count,
                frames_to_read,
                remote_iov + seg_count,
                frames_to_read,
                0ULL);

        if (bytes_read < 0) {
            if (errno == ENOSYS)
                rc = ENOSYS;
            else
                _E("process_vm_readv: %s", strerror(errno));

            goto process_vm_readv_end;
        }

        bytes_total += bytes_read;
        total_length = bytes_total;
        for (seg_count = n_frames-1; seg_count >= 0; --seg_count) {
            if (frame_bytes[seg_count] == bytes_total)
                break;
        }

        if (seg_count < 0) {
            _E("Unknown number of bytes returned by "
                    "process_vm_readv: bytes_read=%zd "
                    "bytes_total=%zd seg_count=%d",
                    bytes_read, bytes_total, seg_count);
            goto process_vm_readv_end;
        }

        if (seg_count == (n_frames-1))
            break;

        ++seg_count;
    }

    rc = 0;

process_vm_readv_end:
    free(local_iov);
    free(remote_iov);
    free(frame_bytes);
    return rc;
}

/*
 * read the file /proc/<pid>/mem
 */
static int copy_memory_proc_mem(int pid, struct mem_data_chunk **frames,
        int n_frames)
{
    int i = 0;
    char fname[32];
    int fd;
    int rc = -1;

    snprintf(fname, sizeof(fname), "/proc/%d/mem", pid);
    if ((fd = open(fname, O_RDONLY)) == -1) {
        _E("Cannot open %s: %s", fname, strerror(errno));
        return -1;
    }

    for (i = 0; i < n_frames; ++i) {
        off_t from = (off_t)(uintptr_t)frames[i]->start;
        char *to = frames[i]->data;
        size_t count = frames[i]->length;

        while (count > 0) {
            ssize_t rd = pread(fd, to, count, from);

            if (rd == -1) {
                _E("pread() at %s:0x%lx (#%d) failed: %s [%d]",
		   fname, from, i, strerror(errno), errno);
                goto proc_mem_end;
            }

            from += rd;
            to += rd;
            count -= rd;
        }

        total_length += frames[i]->length;
    }

    rc = 0;

proc_mem_end:
    close(fd);
    return rc;
}

int copy_memory(int pid, struct mem_data_chunk **frames, int n_frames)
{
    if (!opt_proc_mem) {
        int rc = copy_memory_process_vm_readv(pid, frames, n_frames);
        if (rc == ENOSYS) {
            if (opt_verbose) {
                _E("process_vm_readv is not supported, falling "
                                "back to /proc/pid/mem");
            }
        } else {
            return rc;
        }
    }

    return copy_memory_proc_mem(pid, frames, n_frames);
}

void *get_vdso()
{
    static const char *auxv = "/proc/self/auxv";
    FILE *f;
    long entry[2];

    f = fopen(auxv, "r");
    if (f == NULL) {
        perror(auxv);
        return NULL;
    }

    while (!feof(f)) {
        if (fread(entry, sizeof(entry), 1, f) != 1)
            goto get_vdso_fail;

        if (entry[0] == AT_SYSINFO_EHDR) {
            fclose(f);
            return (void *)entry[1];
        }
    }

get_vdso_fail:
    perror(auxv);
    fclose(f);
    return NULL;
}
